@@ -46,8 +46,6 @@ __author__ = 'nnorwitz@google.com (Neal Norwitz)'
 # * exceptions
 
 
-VISIBILITY_PUBLIC, VISIBILITY_PROTECTED, VISIBILITY_PRIVATE = list(range(3))
-
 FUNCTION_NONE = 0x00
 FUNCTION_SPECIFIER = 0x01
 FUNCTION_VIRTUAL = 0x02
@@ -281,11 +279,12 @@ class Struct(Class):
 class Function(_GenericDeclaration):
 
     def __init__(self, start, end, name, return_type, parameters,
-                 modifiers, templated_types, body, namespace):
+                 specializations, modifiers, templated_types, body, namespace):
         _GenericDeclaration.__init__(self, start, end, name, namespace)
         converter = TypeConverter(namespace)
         self.return_type = converter.create_return_type(return_type)
         self.parameters = converter.to_parameters(parameters)
+        self.specializations = converter.to_type(specializations)
         self.modifiers = modifiers
         self.body = body
         self.templated_types = templated_types
@@ -316,9 +315,10 @@ class Function(_GenericDeclaration):
 class Method(Function):
 
     def __init__(self, start, end, name, in_class, return_type, parameters,
-                 modifiers, templated_types, body, namespace):
+                 specializations, modifiers, templated_types, body, namespace):
         Function.__init__(self, start, end, name, return_type, parameters,
-                          modifiers, templated_types, body, namespace)
+                          specializations, modifiers, templated_types,
+                          body, namespace)
         # TODO(nnorwitz): in_class could also be a namespace which can
         # mess up finding functions properly.
         self.in_class = in_class
@@ -624,7 +624,7 @@ class TypeConverter(object):
 
 class ASTBuilder(object):
 
-    def __init__(self, token_stream, filename, in_class=None, visibility=None,
+    def __init__(self, token_stream, filename, in_class=None,
                  namespace_stack=None, quiet=False):
         if namespace_stack is None:
             namespace_stack = []
@@ -639,7 +639,6 @@ class ASTBuilder(object):
         self.in_class_name_only = None
         if in_class is not None:
             self.in_class_name_only = in_class.split('::')[-1].split('<')[0]
-        self.visibility = visibility
         # Keep the state whether we are currently handling a typedef or not.
         self._handling_typedef = False
         self._handling_const = False
@@ -688,7 +687,8 @@ class ASTBuilder(object):
         if token.token_type == tokenize.NAME:
             if (keywords.is_keyword(token.name) and
                     not keywords.is_builtin_type(token.name)):
-                method = getattr(self, 'handle_' + token.name)
+                method = getattr(self, 'handle_' + token.name, None)
+                assert_parse(method, 'unexpected token: {}'.format(token))
                 return method()
             elif token.name == self.in_class_name_only:
                 # The token name is the same as the class, must be a ctor if
@@ -929,21 +929,22 @@ class ASTBuilder(object):
 
     def _get_method(self, return_type_and_name, modifiers, templated_types,
                     get_paren):
-        template_portion = None
+        specializations = []
         if get_paren:
             token = self._get_next_token()
             assert_parse(token.token_type == tokenize.SYNTAX, token)
             if token.name == '<':
                 # Handle templatized dtors.
-                template_portion = [token]
-                template_portion.extend(self._get_matching_char('<', '>'))
+                specializations = list(self._get_matching_char('<', '>'))
+                del specializations[-1]
                 token = self._get_next_token()
             assert_parse(token.token_type == tokenize.SYNTAX, token)
             assert_parse(token.name == '(', token)
 
         name = return_type_and_name.pop()
-        if (len(return_type_and_name) > 1 and
+        if (len(return_type_and_name) > 2 and
                 return_type_and_name[-1].name == '>' and
+                return_type_and_name[-2].name == 'operator' and
                 (name.name == '>=' or name.name == '>')):
             n = return_type_and_name.pop()
             name = tokenize.Token(tokenize.SYNTAX,
@@ -958,10 +959,16 @@ class ASTBuilder(object):
                                   op.start, name.end)
         # Handle templatized ctors.
         elif name.name == '>':
-            index = 1
-            while return_type_and_name[index].name != '<':
-                index += 1
-            template_portion = return_type_and_name[index:] + [name]
+            count = 1
+            index = len(return_type_and_name)
+            while count and index > 0:
+                index -= 1
+                tok = return_type_and_name[index]
+                if tok.name == '<':
+                    count -= 1
+                elif tok.name == '>':
+                    count += 1
+            specializations = return_type_and_name[index+1:]
             del return_type_and_name[index:]
             name = return_type_and_name.pop()
         elif name.name == ']':
@@ -970,7 +977,6 @@ class ASTBuilder(object):
             name = tokenize.Token(tokenize.NAME, name_seq.name + '[]',
                                   name_seq.start, name.end)
 
-        # TODO(nnorwitz): store template_portion.
         return_type = return_type_and_name
         indices = name
         if return_type:
@@ -1000,7 +1006,8 @@ class ASTBuilder(object):
             if (
                 token.name == 'const' or
                 token.name == 'override' or
-                token.name == 'final'
+                token.name == 'final' or
+                token.name == 'noexcept'
             ):
                 modifiers |= FUNCTION_SPECIFIER
                 token = self._get_next_token()
@@ -1054,11 +1061,10 @@ class ASTBuilder(object):
         # Handle pointer to functions that are really data but look
         # like method declarations.
         if token.name == '(':
-            if parameters[0].name == '*':
+            if parameters[0].name in '*&':
                 # name contains the return type.
+                return_type.append(name)
                 name = parameters.pop()
-                # parameters contains the name of the data.
-                modifiers = [p.name for p in parameters]
                 # Already at the ( to open the parameter list.
                 function_parameters = list(self._get_matching_char('(', ')'))
                 del function_parameters[-1]  # Remove trailing ')'.
@@ -1068,8 +1074,9 @@ class ASTBuilder(object):
                 assert_parse(token.token_type == tokenize.SYNTAX, token)
                 assert_parse(token.name == ';', token)
 
+                types = [t.name for t in return_type]
                 return self._create_variable(indices, name.name, indices.name,
-                                             modifiers, '')
+                                             [], types)
             # At this point, we got something like:
             #  return_type (type::*name_)(params);
             # This is a data member called name_ that is a function pointer.
@@ -1109,11 +1116,11 @@ class ASTBuilder(object):
             return_type, in_class = \
                 self._get_return_type_and_class_name(return_type)
             return Method(indices.start, indices.end, name.name, in_class,
-                          return_type, parameters, modifiers, templated_types,
-                          body, self.namespace_stack)
+                          return_type, parameters, specializations, modifiers,
+                          templated_types, body, self.namespace_stack)
         return Function(indices.start, indices.end, name.name, return_type,
-                        parameters, modifiers, templated_types, body,
-                        self.namespace_stack)
+                        parameters, specializations, modifiers,
+                        templated_types, body, self.namespace_stack)
 
     def _get_return_type_and_class_name(self, token_seq):
         # Splitting the return type from the class name in a method
@@ -1204,7 +1211,7 @@ class ASTBuilder(object):
         assert_parse(token.token_type == tokenize.NAME, token)
         return self._create_variable(token, token.name, name, [], '')
 
-    def _handle_class_and_struct(self, class_type, class_str, visibility):
+    def _handle_class_and_struct(self, class_type, class_str):
         # Special case the handling typedef/aliasing of classes/structs here.
         name_tokens, var_token = self.get_name()
         if name_tokens and not self._handling_typedef:
@@ -1248,15 +1255,13 @@ class ASTBuilder(object):
         else:
             self._add_back_token(var_token)
         self._add_back_tokens(name_tokens)
-        return self._get_class(class_type, visibility, None)
+        return self._get_class(class_type, None)
 
     def handle_class(self):
-        return self._handle_class_and_struct(Class, 'class',
-                                             VISIBILITY_PRIVATE)
+        return self._handle_class_and_struct(Class, 'class')
 
     def handle_struct(self):
-        return self._handle_class_and_struct(Struct, 'struct',
-                                             VISIBILITY_PUBLIC)
+        return self._handle_class_and_struct(Struct, 'struct')
 
     def handle_union(self):
         return self._get_nested_type(Union)
@@ -1291,15 +1296,12 @@ class ASTBuilder(object):
 
     def handle_public(self):
         assert_parse(self.in_class, 'expected to be in a class')
-        self.visibility = VISIBILITY_PUBLIC
 
     def handle_protected(self):
         assert_parse(self.in_class, 'expected to be in a class')
-        self.visibility = VISIBILITY_PROTECTED
 
     def handle_private(self):
         assert_parse(self.in_class, 'expected to be in a class')
-        self.visibility = VISIBILITY_PRIVATE
 
     def handle_friend(self):
         tokens, last = self._get_var_tokens_up_to(False, '(', ';')
@@ -1405,14 +1407,14 @@ class ASTBuilder(object):
         if token.token_type == tokenize.NAME:
             if token.name == 'class':
                 return self._get_class(Class,
-                                       VISIBILITY_PRIVATE,
                                        templated_types)
             elif token.name == 'struct':
                 return self._get_class(Struct,
-                                       VISIBILITY_PUBLIC,
                                        templated_types)
             elif token.name == 'friend':
                 return self.handle_friend()
+            elif token.name == 'template':
+                return self.handle_template()
         self._add_back_token(token)
         tokens, last = self._get_var_tokens_up_to(False, '(', ';')
         tokens.append(last)
@@ -1422,15 +1424,6 @@ class ASTBuilder(object):
         # Must be a variable definition.
         return None
 
-    def handle_true(self):
-        pass  # Nothing to do.
-
-    def handle_false(self):
-        pass  # Nothing to do.
-
-    def handle_asm(self):
-        pass  # Not needed yet.
-
     def _get_bases(self):
         # Get base classes.
         bases = []
@@ -1462,7 +1455,7 @@ class ASTBuilder(object):
             assert_parse(next_token.name == ',', next_token)
         return bases, token
 
-    def _get_class(self, class_type, visibility, templated_types):
+    def _get_class(self, class_type, templated_types):
         class_name = None
         class_token = self._get_next_token()
         if class_token.token_type != tokenize.NAME:
@@ -1514,7 +1507,7 @@ class ASTBuilder(object):
         if token.token_type == tokenize.SYNTAX and token.name == '{':
             name = class_name or '__unamed__'
             ast = ASTBuilder(self.get_scope(), self.filename, name,
-                             visibility, self.namespace_stack,
+                             self.namespace_stack,
                              quiet=self.quiet)
             body = list(ast.generate())
 
